@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1996-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1996-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -72,6 +72,10 @@
 #include <bstring.h>		/* for IRIX, FD_SET calls bzero() */
 #endif
 
+#ifdef APACHE_SSL
+# include <err.h>
+#endif
+
 #ifndef DEFAULT_BUFSIZE
 #define DEFAULT_BUFSIZE (4096)
 #endif
@@ -244,7 +248,23 @@ static int ap_read(BUFF *fb, void *buf, int nbyte)
 #endif
 #ifdef APACHE_SSL
     if(fb->ssl)
+	{
 	rv=SSL_read(fb->ssl,buf,nbyte);
+
+	if(rv < 0)
+	    {
+	    int l;
+
+	    while((l=ERR_get_error()))
+		{
+		char buf[MAX_STRING_LEN];
+
+		ERR_error_string(l,buf);
+		ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,NULL,"%s",
+			     buf);
+		}
+	    }
+	}
     else
 #endif
 	rv = read(fb->fd_in, buf, nbyte);
@@ -264,9 +284,28 @@ static ap_inline int buff_read(BUFF *fb, void *buf, int nbyte)
     }
     else
 	rv = ap_read(fb, buf, nbyte);
+#elif defined(TPF)
+    fd_set fds;
+    struct timeval tv;
+
+    tpf_process_signals();
+    if (fb->flags & B_SOCKET) {
+        alarm(rv = alarm(0));
+        FD_ZERO(&fds);
+        FD_SET(fb->fd_in, &fds);
+        tv.tv_sec = rv+1;
+        tv.tv_usec = 0;
+        rv = ap_select(fb->fd_in + 1, &fds, NULL, NULL, &tv);
+        if (rv < 1) {
+            tpf_process_signals();
+            return(rv);
+        }
+    }
+    rv = ap_read(fb, buf, nbyte);
 #else
     rv = ap_read(fb, buf, nbyte);
 #endif /* WIN32 */
+
     return rv;
 }
 
@@ -284,7 +323,23 @@ static int ap_write(BUFF *fb, const void *buf, int nbyte)
 #endif
 #ifdef APACHE_SSL
     if(fb->ssl)
+	{
 	rv=SSL_write(fb->ssl,buf,nbyte);
+
+	if(rv < 0)
+	    {
+	    int l;
+
+	    while((l=ERR_get_error()))
+		{
+		char buf[MAX_STRING_LEN];
+
+		ERR_error_string(l,buf);
+		ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,NULL,"%s",
+			     buf);
+		}
+	    }
+	}
     else
 #endif
 #if defined (B_SFIO)
@@ -292,7 +347,7 @@ static int ap_write(BUFF *fb, const void *buf, int nbyte)
 #else
 	rv = write(fb->fd, buf, nbyte);
 #endif
-    
+
     return rv;
 }
 
@@ -1230,8 +1285,10 @@ API_EXPORT(int) ap_bwrite(BUFF *fb, const void *buf, int nbyte)
             if (cbuf != NULL)
                 free(cbuf);
             cbuf = malloc(csize = nbyte+HUGE_STRING_LEN);
-            if (cbuf == NULL)
+            if (cbuf == NULL) {
+                fprintf(stderr, "Ouch!  Out of memory in ap_bwrite()!\n");
                 csize = 0;
+            }
         }
         ebcdic2ascii((cbuf) ? cbuf : (void*)buf, buf, nbyte);
         buf = (cbuf) ? cbuf : buf;
@@ -1593,3 +1650,4 @@ API_EXPORT(int) ap_vbprintf(BUFF *fb, const char *fmt, va_list ap)
     }
     return res;
 }
+ 
