@@ -63,13 +63,25 @@
 #include <x509.h>
 #include <pem.h>
 #include <crypto.h>
+#include <rand.h>
 
 #include "httpd.h"
 #include "http_config.h"
 #include "http_log.h"
+#include "http_main.h"
+
+
+#if SSLEAY_VERSION_NUMBER < 0x0920
+# error "Don't use OpenSSL versions less than 0.9.2b, they have a serious security problem!"
+#endif
 
 /* Note that APACHE_SSL_EXPORT_CERTS is defined in buff.h */
 
+/* Temp */
+#define RENEG		FALSE
+#define DEBUG_RENEG	FALSE
+#define DEBUG_SSL_STATE FALSE
+
 #define USE_OLD_ENVIRONMENT	FALSE	/* Set to TRUE to use the old SSL_{CLIENT,SERVER}_I* instead of SSL_{CLIENT,SERVER}_I_* */
 
 #define SSL_IS_OPTIONAL		FALSE	/* Set to TRUE if you want to allow SSL to be disabled on a per-request basis (useful for subrequests) */
@@ -97,7 +109,7 @@
 #define FREE_SESSION	TRUE	/* Don't set this FALSE! */
 
 /* define this to debug the verification stuff */
-#define DEBUG_VERIFY	TRUE
+#define DEBUG_VERIFY	FALSE
 
 #if CACHE_SESSIONS
 # include "gcache.h"
@@ -105,7 +117,7 @@ static const char *s_szCacheServerPort;
 static const char *s_szCacheServerDir;
 static char *s_szCacheServerPath;
 
-#define DEBUG_SESSIONS		0
+# define DEBUG_SESSIONS	FALSE
 
 #endif
 
@@ -155,8 +167,10 @@ typedef struct
     char *szCACertificatePath;
     char *szCACertificateFile;
     char *szReqCiphers;
+#if !RENEG
     int nVerifyDepth;
     VerifyType nVerifyClient;
+#endif
 
     X509 *px509Certificate;
     RSA *prsaKey;
@@ -166,6 +180,8 @@ typedef struct
     BOOL bFakeBasicAuth;
 
     time_t tSessionCacheTimeout;
+
+    array_header *aRandomFiles;
     } SSLConfigRec;
 
 typedef struct
@@ -179,6 +195,10 @@ typedef struct
     BOOL bSSLRequired;
 #if APACHE_SSL_EXPORT_CERTS
     BOOL bExportCerts;
+#endif
+#if RENEG
+    int nVerifyDepth;
+    VerifyType nVerifyClient;
 #endif
     } SSLDirConfigRec;
 
@@ -186,6 +206,9 @@ MODULE_VAR_EXPORT module apache_ssl_module;
 
 static conn_rec *s_pCurrentConnection;
 static pool *s_pPermanentPool;
+#if RENEG
+static SSLDirConfigRec *s_pConfig;
+#endif
 
 static const char six2pr[64+1]=
   "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";
@@ -225,6 +248,241 @@ static void uuencode(char *szTo,const char *szFrom,BOOL bPad)
 	      strlen(szFrom),bPad);
     }
 
+static void SSLLogError(server_rec *s,int nLevel,const char *szErr)
+    {
+    unsigned long l;
+    char buf[MAX_STRING_LEN];
+
+    if(szErr)
+	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",szErr);
+    while((l=ERR_get_error()))
+	{
+	ERR_error_string(l,buf);
+	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",buf);
+	}
+    }
+
+/* for debugging */
+static void SSLLogErrorToFile(FILE *f)
+    {
+    unsigned long l;
+    char buf[MAX_STRING_LEN];
+
+    /* Print out error messages */
+
+    while((l=ERR_get_error()))
+	{
+	ERR_error_string(l,buf);
+	fputs(buf,f);
+	fputc('\n',f);
+	}
+    }
+
+#if DEBUG_SSL_STATE
+
+static void InfoCallback(SSL *s,int where,int ret)
+    {
+    char *str;
+    int w;
+    server_rec *pServer=(server_rec *)SSL_get_ex_data(s,0);
+
+    w=where&~SSL_ST_MASK;
+
+    if(w & SSL_ST_CONNECT)
+	str="SSL_connect";
+    else if(w & SSL_ST_ACCEPT)
+	str="SSL_accept";
+    else
+	str="undefined";
+
+    if(where & SSL_CB_LOOP)
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"%s:%s",
+		     str,SSL_state_string_long(s));
+    else if(where & SSL_CB_ALERT)
+	{
+	str=(where & SSL_CB_READ)?"read":"write";
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		     "SSL3 alert %s:%s:%s",
+		     str,
+		     SSL_alert_type_string_long(ret),
+		     SSL_alert_desc_string_long(ret));
+	}
+    else if(where & SSL_CB_EXIT)
+	{
+	if(ret == 0)
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "%s:failed in %s",
+			 str,SSL_state_string_long(s));
+	else if (ret < 0)
+	    {
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "%s:error in %s",
+			 str,SSL_state_string_long(s));
+	    }
+	}
+    }
+
+#endif /* DEBUG_SSL_STATE */
+
+typedef struct
+    {
+    const char *szFile;
+    int nBytes;
+    } RandomFile;
+
+static void ReadRandomFiles(server_rec *pServer,RandomFile *pRF,int nRF)
+    {
+    int n;
+
+    for(n=0 ; n < nRF ; ++n,++pRF)
+	{
+	int r=RAND_load_file(pRF->szFile,pRF->nBytes);
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		     "Random input %s(%d) -> %d",pRF->szFile,pRF->nBytes,r);
+	}
+    }
+
+#if APACHE_SSL_EXPORT_CERTS
+/* szEnv must have been pool allocated, or be static */
+static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
+    {
+    int n;
+    unsigned char *s,*d;
+
+    n=i2d_X509(xs,NULL);
+    d=s=ap_palloc(p,n);
+    i2d_X509(xs,&d);
+    d=ap_palloc(p,(n*4)/3+2+2);
+    uuencoden(d,s,n,TRUE);
+    ap_table_setn(e,szEnv,(char *)d);
+    }
+#endif
+
+#if SSLEAY_VERSION_NUMBER >= 0x0800
+static int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
+    {
+    X509 *xs=X509_STORE_CTX_get_current_cert(ctx);
+    int depth=X509_STORE_CTX_get_error_depth(ctx);
+    int error=X509_STORE_CTX_get_error(ctx);
+#else
+static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
+				   int error)
+    {
+#endif
+#if RENEG
+    SSLDirConfigRec *pConfig=s_pConfig;
+#else
+    SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
+#endif
+    server_rec *pServer=s_pCurrentConnection->server;
+
+#if DEBUG_VERIFY
+      {
+      char *s;
+
+#if SSLEAY_VERSION_NUMBER < 0x0800
+      s=X509_NAME_oneline(X509_get_subject_name(xs));
+#else
+      s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
+#endif
+
+      if(s == NULL)
+	  {
+	  SSLLogError(pServer,APLOG_ERR,"can't get subject name");
+	  return(0);
+	  }
+
+      ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"depth=%d %s",
+		   depth,s);
+      free(s);
+      }
+#endif
+    
+#if APACHE_SSL_KEEP_CERTS
+    if(depth)
+	{
+	char *t;
+	pool *p=s_pCurrentConnection->client->pool;
+
+	if(!s_pCurrentConnection->client->tableCertChain)
+	    s_pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
+
+	t=ap_psprintf(p,ENV_CERT_CHAIN "%d",depth);
+	AddCertToEnv(p,s_pCurrentConnection->client->tableCertChain,t,xs);
+	}
+#endif
+
+    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
+       || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
+       || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
+       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
+	{
+	char *s;
+
+	if(pConfig->nVerifyClient == VERIFY_OPTIONAL_NO_CA)
+	    {
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "no issuer, returning OK");
+	    return TRUE;
+	    }
+#if SSLEAY_VERSION_NUMBER < 0x0800
+	s=X509_NAME_oneline(X509_get_issuer_name(xs));
+#else
+	s=X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0);
+#endif
+	if(s == NULL)
+	    {
+	    SSLLogError(pServer,APLOG_ERR,"verify error");
+	    return FALSE;
+	    }
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"issuer=%s",
+		     s);
+	free(s);
+	}
+    if(!ok)
+	{
+	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,pServer,
+		      "verify error:num=%d:%s",error,
+		      X509_verify_cert_error_string(error));
+	SSLLogError(s_pCurrentConnection->server,APLOG_ERR,NULL);
+	s_pCurrentConnection->client->szClientX509=NULL;
+	}
+    if(depth >= pConfig->nVerifyDepth)
+	{
+	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,
+		     pServer,"Verify depth exceeded");
+	ok=0;
+	}
+    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		 "verify return:%d",ok);
+    return ok;
+    }
+
+#if RENEG
+static int VerifyFlags(SSLDirConfigRec *pConfig)
+#else
+static int VerifyFlags(SSLConfigRec *pConfig)
+#endif
+    {
+    int nVerify=0;
+
+    switch(pConfig->nVerifyClient)
+	{
+    case VERIFY_REQUIRE:
+	nVerify|=SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
+	break;
+
+    case VERIFY_OPTIONAL:
+    case VERIFY_OPTIONAL_NO_CA:
+	nVerify|=SSL_VERIFY_PEER;
+	break;
+
+    default: break; /* avoid warnings from gcc */
+	}
+
+    return nVerify;
+    }
+
 static int SSLCheckCipher(request_rec *r)
     {
     char *cipher;
@@ -243,6 +501,43 @@ static int SSLCheckCipher(request_rec *r)
     if(pConfig->bDisabled || OPTIONAL_SSL)
 	return DECLINED;
 
+#if RENEG
+# if DEBUG_RENEG
+    fprintf(stderr,"Renegotiated for %s (%x)!\n",r->filename,VerifyFlags(rec));
+# endif
+    if(1)
+	{
+	s_pConfig=rec;
+	SSL_set_verify(r->connection->client->ssl,VerifyFlags(rec),
+		       ApacheSSLVerifyCallback);
+	SSL_renegotiate(r->connection->client->ssl);
+	SSL_do_handshake(r->connection->client->ssl);
+	SSL_write(r->connection->client->ssl,NULL,0);
+	SSL_read(r->connection->client->ssl,NULL,0);
+	}
+
+    if(rec->nVerifyClient != VERIFY_NONE && rec->nVerifyClient != VERIFY_UNSET)
+	{
+	char *s;
+	X509 *xs;
+	xs=SSL_get_peer_certificate(r->connection->client->ssl);
+	if(xs)
+	    {
+	    s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
+	    r->connection->client->szClientX509=ap_pstrdup(r->connection->pool,
+							   s);
+	    free(s);
+	    }
+	}
+
+    if(rec->nVerifyClient == VERIFY_REQUIRE
+       && !r->connection->client->szClientX509)
+	{
+	SSLLogError(r->connection->server,APLOG_ERR,"No client certificate");
+	return FORBIDDEN;
+	}
+#endif
+
     cipher=SSL_get_cipher(r->connection->client->ssl);
 
     if(ap_table_get(rec->tbBannedCiphers,cipher))
@@ -300,22 +595,6 @@ static void ExpandCert(pool *p,table *pEnv,char *szPrefix,char *szCert)
 	}
     }
 
-#if APACHE_SSL_EXPORT_CERTS
-/* szEnv must have been pool allocated, or be static */
-static void AddCertToEnv(pool *p,table *e,char *szEnv,X509 *xs)
-    {
-    int n;
-    unsigned char *s,*d;
-
-    n=i2d_X509(xs,NULL);
-    d=s=ap_palloc(p,n);
-    i2d_X509(xs,&d);
-    d=ap_palloc(p,(n*4)/3+2+2);
-    uuencoden(d,s,n,TRUE);
-    ap_table_setn(e,szEnv,(char *)d);
-    }
-#endif
-
 static struct
     {
     char *szName;
@@ -443,10 +722,9 @@ static int SSLFixups(request_rec *r)
 #if APACHE_SSL_EXPORT_CERTS
 # if !APACHE_SSL_KEEP_CERTS
     pCertStack=SSL_get_peer_cert_chain(r->connection->client->ssl);
-    /* XXX: When this works, check whether we need to start at 1, or add an offset */
-    for(n=1 ; pCertStack && n < sk_num(pCertStack) ; ++n)
+    for(n=0 ; pCertStack && n < sk_num(pCertStack) ; ++n)
 	{
-	t=ap_psprintf(r->pool,ENV_CERT_CHAIN "%d",n);
+	t=ap_psprintf(r->pool,ENV_CERT_CHAIN "%d",n+1);
 	AddCertToEnv(r->connection->client->pool,e,t,(X509 *)sk_value(pCertStack,n));
 	}
 # else
@@ -460,36 +738,6 @@ static int SSLFixups(request_rec *r)
     return DECLINED;
     }
 
-static void SSLLogError(server_rec *s,int nLevel,const char *szErr)
-    {
-    unsigned long l;
-    char buf[MAX_STRING_LEN];
-
-    if(szErr)
-	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",szErr);
-    while((l=ERR_get_error()))
-	{
-	ERR_error_string(l,buf);
-	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",buf);
-	}
-    }
-
-/* for debugging */
-static void SSLLogErrorToFile(FILE *f)
-    {
-    unsigned long l;
-    char buf[MAX_STRING_LEN];
-
-    /* Print out error messages */
-
-    while((l=ERR_get_error()))
-	{
-	ERR_error_string(l,buf);
-	fputs(buf,f);
-	fputc('\n',f);
-	}
-    }
-
 static void *CreateSSLServerConfig(pool *p,server_rec *s)
     {
     SSLConfigRec *rec=ap_pcalloc(p,sizeof(SSLConfigRec));
@@ -500,13 +748,16 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     rec->szCACertificatePath=NULL;
     rec->szCACertificateFile=NULL;
     rec->szReqCiphers=NULL;
+#if !RENEG
     rec->nVerifyDepth=0;
     rec->nVerifyClient=VERIFY_UNSET;	/* Will become VERIFY_NONE */
+#endif
     rec->px509Certificate=NULL;
     rec->prsaKey=NULL;
     rec->pSSLCtx=NULL;
     rec->bFakeBasicAuth=UNSET;		/* Will become FALSE */
     rec->tSessionCacheTimeout=0;
+    rec->aRandomFiles=ap_make_array(p,1,sizeof(RandomFile));
 
     return rec;
     }
@@ -527,10 +778,13 @@ static void *MergeSSLServerConfig(pool *p,void *basev,void *addv)
     MergeString(szKeyFile);
     MergeString(szCACertificatePath);
     MergeString(szCACertificateFile);
+#if !RENEG
     MergeInt(nVerifyDepth);
     Merge(nVerifyClient,VERIFY_UNSET);
+#endif
     MergeBool(bFakeBasicAuth);
     MergeInt(tSessionCacheTimeout);
+    new->aRandomFiles=ap_append_arrays(p,base->aRandomFiles,add->aRandomFiles);
 
     return new;
     }
@@ -545,6 +799,10 @@ static void *CreateSSLDirConfig(pool *p,char *dummy)
 #if APACHE_SSL_EXPORT_CERTS
     rec->bExportCerts=FALSE;
 #endif
+#if RENEG
+    rec->nVerifyDepth=0;
+    rec->nVerifyClient=VERIFY_UNSET;	/* Won't become VERIFY_NONE */
+#endif
 
     return rec;
     }
@@ -564,6 +822,11 @@ static void *MergeSSLDirConfig(pool *p,void *basev,void *addv)
 #if APACHE_SSL_EXPORT_CERTS
     new->bExportCerts=base->bExportCerts || add->bExportCerts;
 #endif
+#if RENEG
+    MergeInt(nVerifyDepth);
+    Merge(nVerifyClient,VERIFY_UNSET);
+#endif
+
     return new;
     }
 
@@ -594,9 +857,9 @@ static void InitSSL(server_rec *s)
 #endif
     }
 
-#if SSLEAY_VERSION_NUMBER >= 0x0800
-/* FIXME: This is an expensive operation which should probably be done before
-forking */
+#if SSLEAY_VERSION_NUMBER >= 0x0800 && SSLEAY_VERSION_NUMBER < 0x0920
+/* FIXME: This ought to generate a new key from time to time - why make the
+Feds' life easy? */
 static RSA *TmpRSACallback(SSL *pSSL,int nExport)
     {
     static RSA *pRSA=NULL;
@@ -611,121 +874,26 @@ static RSA *TmpRSACallback(SSL *pSSL,int nExport)
     }
 #endif
 
-#if SSLEAY_VERSION_NUMBER >= 0x0800
-static int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
+#if SSLEAY_VERSION_NUMBER >= 0x0920
+/* FIXME: This ought to generate a new key from time to time - why make the
+Feds' life easy? */
+static RSA *TmpRSACallback(SSL *pSSL,int nExport,int nKeyLength)
     {
-    X509 *xs=X509_STORE_CTX_get_current_cert(ctx);
-    int depth=X509_STORE_CTX_get_error_depth(ctx);
-    int error=X509_STORE_CTX_get_error(ctx);
-#else
-static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
-				   int error)
-    {
-#endif
-    SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
-    server_rec *pServer=s_pCurrentConnection->server;
-    pool *p=s_pCurrentConnection->client->pool;
+    static RSA *pRSA512=NULL;
+    static RSA *pRSA1024=NULL;
 
-#if DEBUG_VERIFY
-      {
-      char *s;
-
-#if SSLEAY_VERSION_NUMBER < 0x0800
-      s=X509_NAME_oneline(X509_get_subject_name(xs));
-#else
-      s=X509_NAME_oneline(X509_get_subject_name(xs),NULL,0);
-#endif
-
-      if(s == NULL)
-	  {
-	  SSLLogError(pServer,APLOG_ERR,"can't get subject name");
-	  return(0);
-	  }
-
-      ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"depth=%d %s",
-		   depth,s);
-      free(s);
-      }
-#endif
+    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,SSL_get_ex_data(pSSL,0),
+		 "Generating %d bit key",nKeyLength);
+    assert(nExport);  /* Otherwise it was a sign-only key and we should set our own keylength */
+    assert(nKeyLength == 512 || nKeyLength == 1024);
+    if (pRSA512 == NULL && nKeyLength == 512)
+	pRSA512=RSA_generate_key(512,RSA_F4,NULL,NULL);
+    if (pRSA1024 == NULL && nKeyLength == 1024)
+	pRSA1024=RSA_generate_key(1024,RSA_F4,NULL,NULL);
     
-#if APACHE_SSL_KEEP_CERTS
-    if(depth)
-	{
-	char *t;
-
-	if(!s_pCurrentConnection->client->tableCertChain)
-	    s_pCurrentConnection->client->tableCertChain=ap_make_table(p,5);
-
-	t=ap_psprintf(p,ENV_CERT_CHAIN "%d",depth);
-	AddCertToEnv(p,s_pCurrentConnection->client->tableCertChain,t,xs);
-	}
-#endif
-
-    if(error == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY
-       || error == X509_V_ERR_UNABLE_TO_VERIFY_LEAF_SIGNATURE
-       || error == X509_V_ERR_DEPTH_ZERO_SELF_SIGNED_CERT
-       || error == X509_V_ERR_SELF_SIGNED_CERT_IN_CHAIN)
-	{
-	char *s;
-
-	if(pConfig->nVerifyClient == VERIFY_OPTIONAL_NO_CA)
-	    {
-	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
-			 "no issuer, returning OK");
-	    return TRUE;
-	    }
-#if SSLEAY_VERSION_NUMBER < 0x0800
-	s=X509_NAME_oneline(X509_get_issuer_name(xs));
-#else
-	s=X509_NAME_oneline(X509_get_issuer_name(xs),NULL,0);
-#endif
-	if(s == NULL)
-	    {
-	    SSLLogError(pServer,APLOG_ERR,"verify error");
-	    return FALSE;
-	    }
-	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"issuer=%s",
-		     s);
-	free(s);
-	}
-    if(!ok)
-	{
-	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,pServer,
-		      "verify error:num=%d:%s",error,
-		      X509_verify_cert_error_string(error));
-	SSLLogError(s_pCurrentConnection->server,APLOG_ERR,NULL);
-	s_pCurrentConnection->client->szClientX509=NULL;
-	}
-    if(depth >= pConfig->nVerifyDepth)
-	{
-	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,
-		     pServer,"Verify depth exceeded");
-	ok=0;
-	}
-    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
-		 "verify return:%d",ok);
-    return ok;
-    }
-
-static int VerifyFlags(SSLConfigRec *pConfig)
-    {
-    int nVerify=0;
-
-    switch(pConfig->nVerifyClient)
-	{
-    case VERIFY_REQUIRE:
-	nVerify|=SSL_VERIFY_PEER|SSL_VERIFY_FAIL_IF_NO_PEER_CERT;
-	break;
-
-    case VERIFY_OPTIONAL:
-    case VERIFY_OPTIONAL_NO_CA:
-	nVerify|=SSL_VERIFY_PEER;
-	break;
-
-    default: break; /* avoid warnings from gcc */
-	}
-    return nVerify;
+    return nKeyLength == 512 ? pRSA512 : pRSA1024;
     }
+#endif
 
 #if CACHE_SESSIONS
 
@@ -769,9 +937,7 @@ static uchar *CacheFind(uchar *aucKey,int nKey,int *pnData,BOOL *pbLocal)
     return t;
     }
 
-static uchar *CacheFindS(char *aucKey,int nKey,int *pnData,BOOL *pbLocal)
-    { return CacheFind((uchar *)aucKey,nKey,pnData,pbLocal); }
-
+#if APACHE_SSL_KEEP_CERTS
 static uchar *MakeCertKey(Cache *pCache,const char *szKey,int *pnLength)
     {
     uchar *buf=malloc(*pnLength=pCache->nKey+strlen(szKey));
@@ -797,7 +963,9 @@ static int SendCertToServer(void *extra,const char *szKey,const char *szValue)
     return 1;
     }
 
-#if APACHE_SSL_EXPORT_CERTS
+static uchar *CacheFindS(char *aucKey,int nKey,int *pnData,BOOL *pbLocal)
+    { return CacheFind((uchar *)aucKey,nKey,pnData,pbLocal); }
+
 static void SendCertChainToServer(Cache *pCache)
     {
     if(!s_pCurrentConnection->client->tableCertChain)
@@ -835,7 +1003,7 @@ static void GetCertChainFromServer(uchar *aucSessionID,int nLength,
 		     buf+nLength,(char *)t);
 	}
     }
-#endif
+#endif /* APACHE_SSL_KEEP_CERTS */
 
 static void SendSessionToServer(SSL_SESSION *pSession)
     {
@@ -867,7 +1035,7 @@ static void SendSessionToServer(SSL_SESSION *pSession)
     p->nData=nLength;
 
     GlobalCacheAdd(p->aucKey,p->nKey,p->aucData,p->nData,tExpiresAt);
-#if APACHE_SSL_EXPORT_CERTS
+#if APACHE_SSL_KEEP_CERTS
     SendCertChainToServer(p);
 #endif
     }
@@ -879,7 +1047,7 @@ static SSL_SESSION *GetSessionFromServer(uchar *aucSessionID,int nLength)
     int nData;
     BOOL bLocal=FALSE;
 
-#if APACHE_SSL_EXPORT_CERTS
+#if APACHE_SSL_KEEP_CERTS
     GetCertChainFromServer(aucSessionID,nLength,bLocal);
 #endif
 
@@ -983,15 +1151,20 @@ static void GetPrivateKey(server_rec *s,SSLConfigRec *pConfig,
 		  (const char *)pConfig->prsaKey);
     }
 
+#if SSLEAY_VERSION_NUMBER < 0x0920 || RENEG
 static int X509NameCompare(X509_NAME **a,X509_NAME **b)
     {
-    return(X509_NAME_cmp(*a,*b));
+    return X509_NAME_cmp(*a,*b);
     }
+#endif
 
 static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
     {
     FILE *f;
     char szPath[MAX_STRING_LEN];
+#if SSLEAY_VERSION_NUMBER >= 0x0920 && !RENEG
+    STACK *stackCAs=NULL;
+#endif
 
     if(pConfig->px509Certificate)
 	{
@@ -1005,7 +1178,12 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
     pConfig->pSSLCtx=SSL_CTX_new();
 #else
     pConfig->pSSLCtx=SSL_CTX_new(SSLv23_server_method());
+# if RENEG
+//    s_nVerifyClient=VERIFY_OPTIONAL_NO_CA;
+    SSL_CTX_set_verify(pConfig->pSSLCtx,0,ApacheSSLVerifyCallback);
+# else
     SSL_CTX_set_verify(pConfig->pSSLCtx,VerifyFlags(pConfig),ApacheSSLVerifyCallback);
+# endif
 #endif
 
 #if CACHE_SESSIONS
@@ -1079,7 +1257,34 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 
 #if SSLEAY_VERSION_NUMBER >= 0x0800
     SSL_CTX_set_tmp_rsa_callback(pConfig->pSSLCtx,TmpRSACallback);
+#endif
 
+#if SSLEAY_VERSION_NUMBER >= 0x0920 && !RENEG
+    if(pConfig->nVerifyClient != VERIFY_OPTIONAL_NO_CA)
+	{
+	if(pConfig->szCACertificateFile || pConfig->szCACertificatePath)
+	    stackCAs=sk_new(NULL);
+
+	if(pConfig->szCACertificateFile
+	   && !SSL_add_file_cert_subjects_to_stack(stackCAs,
+					       pConfig->szCACertificateFile))
+	    {
+	    SSLLogError(s,APLOG_CRIT,"error reading CA certs");
+	    exit(1);
+	    }
+
+	if(pConfig->szCACertificatePath
+	   && !SSL_add_dir_cert_subjects_to_stack(stackCAs,
+						  pConfig->szCACertificatePath))
+	    {
+	    SSLLogError(s,APLOG_CRIT,"error reading CA certs");
+	    exit(1);
+	    }
+
+	if(stackCAs && sk_num(stackCAs) > 0)
+	    SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,stackCAs);
+	}
+#elif SSLEAY_VERSION_NUMBER >= 0x0800
     if(pConfig->szCACertificateFile)
 	SSL_CTX_set_client_CA_list(pConfig->pSSLCtx,
 				   SSL_load_client_CA_file(pConfig->szCACertificateFile));
@@ -1100,7 +1305,7 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 	    exit(1);
 	    }
 
-	while(dstruct=readdir(d))
+	while((dstruct=readdir(d)))
 	    {
 	    STACK *s;
 	    int n;
@@ -1111,8 +1316,6 @@ static void GetCertificateAndKey(server_rec *s,pool *p,SSLConfigRec *pConfig)
 	    s=SSL_load_client_CA_file(szFile);
 	    for(n=0 ; s && n < sk_num(s) ; ++n)
 		{
-		printf("Reading %s\n",
-		       X509_NAME_oneline((X509_NAME *)sk_value(s,n),NULL,0));
 		if(sk_find(stackCAs,sk_value(s,n)) >= 0)
 		    continue;
 	        sk_push(stackCAs,sk_value(s,n));
@@ -1199,8 +1402,10 @@ static void InitSSLServer(server_rec *s,pool *p)
 
 	if(pConfig->bDisabled == UNSET)
 	    pConfig->bDisabled=FALSE;
+#if !RENEG
 	if(pConfig->nVerifyClient == VERIFY_UNSET)
 	    pConfig->nVerifyClient=VERIFY_NONE;
+#endif
 	if(pConfig->bFakeBasicAuth == UNSET)
 	    pConfig->bFakeBasicAuth=FALSE;
 
@@ -1233,6 +1438,7 @@ static void InitSSLServer(server_rec *s,pool *p)
 			 s->server_hostname,s->port);
 	    exit(1);
 	    }
+#if !RENEG
 	if(pConfig->nVerifyClient < 0 || pConfig->nVerifyClient > VERIFY_OPTIONAL_NO_CA)
 	    {
 	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
@@ -1240,6 +1446,7 @@ static void InitSSLServer(server_rec *s,pool *p)
 			 pConfig->nVerifyClient);
 	    exit(1);
 	    }
+#endif
 	GetCertificateAndKey(s,p,pConfig);
 	}
 #if CACHE_SESSIONS
@@ -1278,15 +1485,15 @@ static int FakeBasicAuth(request_rec *r)
     if(!r->connection->client->szClientX509)
 	return DECLINED;
     /*
-       Fake a password - which one would be immaterial, as, it seems, an empty
-       password in the users file would match ALL incoming passwords, if only we
-       were using the standard crypt library routine. Unfortunately, SSLeay
-       "fixes" a "bug" in crypt and thus prevents blank passwords from working.
-       (IMHO what they really fix is a bug in the users of the code - failing to
-       program correctly for shadow passwords).
-       We need, therefore, to provide a password. This password can be matched by
-       adding the string "xxj31ZMTZzkVA" as the password in the user file.
-*/
+      Fake a password - which one would be immaterial, as, it seems, an empty
+      password in the users file would match ALL incoming passwords, if only we
+      were using the standard crypt library routine. Unfortunately, SSLeay
+      "fixes" a "bug" in crypt and thus prevents blank passwords from working.
+      (IMHO what they really fix is a bug in the users of the code - failing to
+      program correctly for shadow passwords).
+      We need, therefore, to provide a password. This password can be matched
+      by adding the string "xxj31ZMTZzkVA" as the password in the user file.
+      */
     ap_snprintf(b1,sizeof b1,"%s:password",
 		r->connection->client->szClientX509);
     uuencode(b2,b1,FALSE);
@@ -1357,6 +1564,9 @@ static void CloseSSL(void *_conn)
     SSL_free(ssl);
     /*    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,conn->server,
 		 "Closed");*/
+#if RENEG
+    s_pCurrentConnection=NULL;
+#endif
     }
 #endif
 
@@ -1366,23 +1576,42 @@ int ApacheSSLSetupConnection(conn_rec * conn)
     BUFF *fb=conn->client;
     SSLConfigRec *pConfig=ap_get_module_config(srvr->module_config,
 					       &apache_ssl_module);
+    int r;
 
     if(pConfig->bDisabled)
 	{
 	fb->ssl=NULL;
 	return TRUE;
 	}
-    
+
+    ap_conn_timeout("SSL accept timeout",conn);
+
+    ReadRandomFiles(srvr,(RandomFile *)pConfig->aRandomFiles->elts,
+		    pConfig->aRandomFiles->nelts);
+
     s_pCurrentConnection=conn;
 
 #if APACHE_SSL_KEEP_CERTS
     fb->tableCertChain=NULL;
 #endif
+#if SSLEAY_VERSION_NUMBER >= 0x0920
+    /* Use our current configuration as the session ID context. Do this by
+       actually using its address, which should be invariant */
     fb->ssl=SSL_new(pConfig->pSSLCtx);
+    r=SSL_set_session_id_context(fb->ssl,(unsigned char *)&pConfig,
+				 sizeof pConfig);
+    ap_assert(r);
+#else
+    fb->ssl=SSL_new(pConfig->pSSLCtx);
+#endif
     SSL_set_fd(fb->ssl,fb->fd);
 #if FREE_SESSION
     ap_register_cleanup(conn->pool,conn,CloseSSL,CloseSSL);
 #endif
+#if DEBUG_SSL_STATE
+    SSL_set_info_callback(fb->ssl,InfoCallback);
+    SSL_set_ex_data(fb->ssl,0,srvr);
+#endif
 
     ApacheSSLSetupVerify(conn);
 
@@ -1413,6 +1642,7 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	    return FALSE;
 	    }
 
+#if !RENEG
 	if(pConfig->nVerifyClient != VERIFY_NONE)
 	    {
 	    char *s;
@@ -1431,14 +1661,17 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	    SSLLogError(conn->server,APLOG_ERR,"No client certificate");
 	    return 0;
 	    }
+#endif
 	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,srvr,"CIPHER is %s",
 		     SSL_get_cipher(conn->client->ssl));
 	}
 
+#if !RENEG
     /* This should be safe.... so I'll use it */
     SSL_set_read_ahead(fb->ssl,1);
 
     s_pCurrentConnection=NULL;
+#endif
 
     return TRUE;
     }
@@ -1448,6 +1681,17 @@ static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
     return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
     }
 
+#if RENEG
+static const char *SSLVerifyClient(cmd_parms *cmd,SSLDirConfigRec *rec,
+				   char *level)
+    {
+    rec->nVerifyClient=atoi(level);
+    if(rec->nVerifyClient < 0 || rec->nVerifyClient > 3)
+	return "Bad value for SSLVerifyClient";
+    return NULL;
+    }
+#endif
+
 static const char *SSLRequireCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
 				    char *cipher)
     {
@@ -1508,9 +1752,34 @@ static const char *SSLCacheServerDir(cmd_parms *cmd,char *struct_ptr,
     }
 #endif
 
-static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
+static const char *SSLRandomFile(cmd_parms *cmd,char *struct_ptr,char *arg1,
+				 char *arg2)
+    {
+    RandomFile rf;
+
+    rf.szFile=arg1;
+    rf.nBytes=atoi(arg2);
+    ReadRandomFiles(cmd->server,&rf,1);
+
+    return NULL;
+    }
+
+static const char *SSLRandomFilePerConnection(cmd_parms *cmd,char *struct_ptr,
+					      char *arg1,char *arg2)
+    {
+    SSLConfigRec *pConfig=ap_get_module_config(cmd->server->module_config,
+						  &apache_ssl_module);
+    RandomFile *prf=ap_push_array(pConfig->aRandomFiles);
+
+    prf->szFile=arg1;
+    prf->nBytes=atoi(arg2);
+
+    return NULL;
+    }
+
+static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,
+					   char *arg)
     {
-    /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1519,9 +1788,18 @@ static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char
     return NULL;
     }
 
-static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
+#if RENEG
+static const char *set_dir_int_slot(cmd_parms *cmd,char *pStruct,char *szValue)
+    {
+    int offset=(int)(long)cmd->info;
+    *(int *)(pStruct+offset)=atoi(szValue);
+    return NULL;
+    }
+#endif
+
+static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,
+					char *arg)
     {
-  /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1532,7 +1810,6 @@ static const char *set_server_int_slot (cmd_parms *cmd,char *struct_ptr,char *ar
 
 static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
     {
-  /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1543,7 +1820,6 @@ static const char *set_server_bool_slot (cmd_parms *cmd,char *struct_ptr)
 
 static const char *set_server_bool_slot_false (cmd_parms *cmd,char *struct_ptr)
     {
-  /* This one should be pretty generic... */
     char *pConfig=ap_get_module_config(cmd->server->module_config,
 				       &apache_ssl_module);
   
@@ -1572,12 +1848,20 @@ command_rec ssl_cmds[]=
     { "SSLCACertificateFile",set_server_string_slot,
       (void *)XtOffsetOf(SSLConfigRec,szCACertificateFile),RSRC_CONF,TAKE1,
       "CA Certificate file (taken from SSL_CERT_FILE if absent)" },
+#if RENEG
+    { "SSLVerifyDepth",set_dir_int_slot,
+      (void *)XtOffsetOf(SSLDirConfigRec,nVerifyDepth),OR_FILEINFO,TAKE1,
+      "Verify depth (default 0)" },
+    { "SSLVerifyClient",SSLVerifyClient,NULL,OR_FILEINFO,TAKE1,
+      "Verify client (0=no,1=optional,2=required,3=required (CA optional)" },
+#else
     { "SSLVerifyDepth",set_server_int_slot,
       (void *)XtOffsetOf(SSLConfigRec,nVerifyDepth),RSRC_CONF,TAKE1,
       "Verify depth (default 0)" },
     { "SSLVerifyClient",set_server_int_slot,
       (void *)XtOffsetOf(SSLConfigRec,nVerifyClient),RSRC_CONF,TAKE1,
       "Verify client (0=no,1=optional,2=required" },
+#endif
     { "SSLFakeBasicAuth",set_server_bool_slot,
       (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
       "Translate client X509 into a user name" },
@@ -1606,6 +1890,11 @@ command_rec ssl_cmds[]=
     { "SSLCacheServerRunDir",SSLCacheServerDir,NULL,RSRC_CONF,TAKE1,
       "Path for the server to run in (in case of coredumps)" },
 #endif
+    { "SSLRandomFile",SSLRandomFile,NULL,RSRC_CONF,TAKE2,
+      "A file to read randomness from and a maximum number of bytes to read" },
+    { "SSLRandomFilePerConnection", SSLRandomFilePerConnection,NULL,RSRC_CONF,
+      TAKE2,
+      "A file to read randomness from and a maximim number of bytes to read on each request" },
     { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
       "Defunct SSL log file - DO NOT USE" },
     { NULL },
