@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
+ * Copyright (c) 1995-1999 The Apache Group.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -106,8 +106,8 @@
  *   CustomLog   logs/referer  "%{referer}i -> %U"
  *   CustomLog   logs/agent    "%{user-agent}i"
  *
- * Except: no RefererIgnore functionality
- *         logs '-' if no Referer or User-Agent instead of nothing
+ * RefererIgnore functionality can be obtained with conditional
+ * logging (SetEnvIf and CustomLog ... env=!VAR).
  *
  * But using this method allows much easier modification of the
  * log format, e.g. to log hosts along with UA:
@@ -147,7 +147,8 @@
  * %...T:  the time taken to serve the request, in seconds.
  * %...u:  remote user (from auth; may be bogus if return status (%s) is 401)
  * %...U:  the URL path requested.
- * %...v:  the name of the server (i.e. which virtual host?)
+ * %...v:  the configured name of the server (i.e. which virtual host?)
+ * %...V:  the server name according to the UseCanonicalName setting
  *
  * The '...' can be nothing at all (e.g. "%h %u %r %s %b"), or it can
  * indicate conditions for inclusion of the item (which will cause it
@@ -248,6 +249,7 @@ typedef struct {
     char *format_string;
     array_header *format;
     int log_fd;
+    char *condition_var;
 #ifdef BUFFERED_LOGS
     int outcnt;
     char outbuf[LOG_BUFSIZE];
@@ -420,12 +422,21 @@ static const char *log_request_duration(request_rec *r, char *a)
  */
 static const char *log_virtual_host(request_rec *r, char *a)
 {
-    return ap_get_server_name(r);
+    return r->server->server_hostname;
 }
 
 static const char *log_server_port(request_rec *r, char *a)
 {
-    return ap_psprintf(r->pool, "%u", ap_get_server_port(r));
+    return ap_psprintf(r->pool, "%u",
+	r->server->port ? r->server->port : ap_default_port(r));
+}
+
+/* This respects the setting of UseCanonicalName so that
+ * the dynamic mass virtual hosting trick works better.
+ */
+static const char *log_server_name(request_rec *r, char *a)
+{
+    return ap_get_server_name(r);
 }
 
 static const char *log_child_pid(request_rec *r, char *a)
@@ -493,6 +504,9 @@ static struct log_item_list {
     {
         'e', log_env_var, 0
     },
+    {
+        'V', log_server_name, 0
+    },
     {
         'v', log_virtual_host, 0
     },
@@ -522,30 +536,61 @@ static struct log_item_list *find_log_func(char k)
     return NULL;
 }
 
-static char *log_format_substring(pool *p, const char *start,
-                                  const char *end)
-{
-    char *res = ap_palloc(p, end - start + 1);
-
-    strncpy(res, start, end - start);
-    res[end - start] = '\0';
-    return res;
-}
-
 static char *parse_log_misc_string(pool *p, log_format_item *it,
                                    const char **sa)
 {
-    const char *s = *sa;
+    const char *s;
+    char *d;
 
     it->func = constant_item;
     it->conditions = NULL;
 
+    s = *sa;
     while (*s && *s != '%') {
-        ++s;
+	s++;
     }
-    it->arg = log_format_substring(p, *sa, s);
-    *sa = s;
+    /*
+     * This might allocate a few chars extra if there's a backslash
+     * escape in the format string.
+     */
+    it->arg = ap_palloc(p, s - *sa + 1);
 
+    d = it->arg;
+    s = *sa;
+    while (*s && *s != '%') {
+	if (*s != '\\') {
+	    *d++ = *s++;
+	}
+	else {
+	    s++;
+	    switch (*s) {
+	    case '\\':
+		*d++ = '\\';
+		s++;
+		break;
+	    case 'n':
+		*d++ = '\n';
+		s++;
+		break;
+	    case 't':	
+		*d++ = '\t';
+		s++;
+		break;
+	    default:
+		/* copy verbatim */
+		*d++ = '\\';
+		/*
+		 * Allow the loop to deal with this *s in the normal
+		 * fashion so that it handles end of string etc.
+		 * properly.
+		 */
+		break;
+	    }
+	}
+    }
+    *d = '\0';
+
+    *sa = s;
     return NULL;
 }
 
@@ -773,11 +818,30 @@ static int config_log_transaction(request_rec *r, config_log_state *cls,
     int i;
     int len = 0;
     array_header *format;
+    char *envar;
 
     if (cls->fname == NULL) {
         return DECLINED;
     }
 
+    /*
+     * See if we've got any conditional envariable-controlled logging decisions
+     * to make.
+     */
+    if (cls->condition_var != NULL) {
+	envar = cls->condition_var;
+	if (*envar != '!') {
+	    if (ap_table_get(r->subprocess_env, envar) == NULL) {
+		return DECLINED;
+	    }
+	}
+	else {
+	    if (ap_table_get(r->subprocess_env, &envar[1]) != NULL) {
+		return DECLINED;
+	    }
+	}
+    }
+
     format = cls->format ? cls->format : default_format;
 
     strs = ap_palloc(r->pool, sizeof(char *) * (format->nelts));
@@ -836,10 +900,13 @@ static int config_log_transaction(request_rec *r, config_log_state *cls,
 static int multi_log_transaction(request_rec *r)
 {
     multi_log_state *mls = ap_get_module_config(r->server->module_config,
-                                             &config_log_module);
+						&config_log_module);
     config_log_state *clsarray;
     int i;
 
+    /*
+     * Log this transaction..
+     */
     if (mls->config_logs->nelts) {
         clsarray = (config_log_state *) mls->config_logs->elts;
         for (i = 0; i < mls->config_logs->nelts; ++i) {
@@ -867,8 +934,9 @@ static int multi_log_transaction(request_rec *r)
 
 static void *make_config_log_state(pool *p, server_rec *s)
 {
-    multi_log_state *mls = (multi_log_state *) ap_palloc(p, sizeof(multi_log_state));
+    multi_log_state *mls;
 
+    mls = (multi_log_state *) ap_palloc(p, sizeof(multi_log_state));
     mls->config_logs = ap_make_array(p, 1, sizeof(config_log_state));
     mls->default_format_string = NULL;
     mls->default_format = NULL;
@@ -908,7 +976,7 @@ static const char *log_format(cmd_parms *cmd, void *dummy, char *fmt,
 {
     const char *err_string = NULL;
     multi_log_state *mls = ap_get_module_config(cmd->server->module_config,
-                                             &config_log_module);
+						&config_log_module);
 
     /*
      * If we were given two arguments, the second is a name to be given to the
@@ -928,18 +996,31 @@ static const char *log_format(cmd_parms *cmd, void *dummy, char *fmt,
     return err_string;
 }
 
+
 static const char *add_custom_log(cmd_parms *cmd, void *dummy, char *fn,
-                                  char *fmt)
+                                  char *fmt, char *envclause)
 {
     const char *err_string = NULL;
     multi_log_state *mls = ap_get_module_config(cmd->server->module_config,
-                                             &config_log_module);
+						&config_log_module);
     config_log_state *cls;
 
     cls = (config_log_state *) ap_push_array(mls->config_logs);
+    cls->condition_var = NULL;
+    if (envclause != NULL) {
+	if (strncasecmp(envclause, "env=", 4) != 0) {
+	    return "error in condition clause";
+	}
+	if ((envclause[4] == '\0')
+	    || ((envclause[4] == '!') && (envclause[5] == '\0'))) {
+	    return "missing environment variable name";
+	}
+	cls->condition_var = ap_pstrdup(cmd->pool, &envclause[4]);
+    }
+
     cls->fname = fn;
     cls->format_string = fmt;
-    if (!fmt) {
+    if (fmt == NULL) {
         cls->format = NULL;
     }
     else {
@@ -952,18 +1033,19 @@ static const char *add_custom_log(cmd_parms *cmd, void *dummy, char *fn,
 
 static const char *set_transfer_log(cmd_parms *cmd, void *dummy, char *fn)
 {
-    return add_custom_log(cmd, dummy, fn, NULL);
+    return add_custom_log(cmd, dummy, fn, NULL, NULL);
 }
 
 static const char *set_cookie_log(cmd_parms *cmd, void *dummy, char *fn)
 {
-    return add_custom_log(cmd, dummy, fn, "%{Cookie}n \"%r\" %t");
+    return add_custom_log(cmd, dummy, fn, "%{Cookie}n \"%r\" %t", NULL);
 }
 
 static const command_rec config_log_cmds[] =
 {
-    {"CustomLog", add_custom_log, NULL, RSRC_CONF, TAKE2,
-     "a file name and a custom log format string or format name"},
+    {"CustomLog", add_custom_log, NULL, RSRC_CONF, TAKE23,
+     "a file name, a custom log format string or format name, "
+     "and an optional \"env=\" clause (see docs)"},
     {"TransferLog", set_transfer_log, NULL, RSRC_CONF, TAKE1,
      "the filename of the access log"},
     {"LogFormat", log_format, NULL, RSRC_CONF, TAKE12,
@@ -998,7 +1080,7 @@ static config_log_state *open_config_log(server_rec *s, pool *p,
         char *fname = ap_server_root_relative(p, cls->fname);
         if ((cls->log_fd = ap_popenf(p, fname, xfer_flags, xfer_mode)) < 0) {
             ap_log_error(APLOG_MARK, APLOG_ERR, s,
-                         "httpd: could not open transfer log file %s.", fname);
+                         "could not open transfer log file %s.", fname);
             exit(1);
         }
     }
